/**
 * File:  MPR121_drv.c
 *
 * I/O driver functions for NXP/Freescale MPR121 capacitive touch sensor device.
 * This module should not contain application-specific code.
 *
 * This version assumes MPR121 is connected to the MCU I2C1 peripheral.
 */
#include "MPR121_drv.h"
#include "I2C_drv.h"


/*^
 * Function:     Write byte value into a specified register of the MPR121.
 *               The data written is *not* verified by a read-back operation.
 *               I2C slave address is 0x5A; Write control byte is 0xB4.
 *
 * Entry args:   regNum  = register number (address)
 *               regData = data byte value to be written
 *
 * Return val:   ERROR (-1) if an I2C comm fault was detected; otherwise zero (0).
 *
 * Note:         Register writes can be effected only with the MPR121 in Stop Mode,
 *               except for the ECR ($5E) and GPIO/LED related registers.
 *               To put MPR121 in Stop Mode, write $00 to the ECR (reg $5E).
 *               To put MPR121 in Run Mode, write $0n to the ECR (for n touch inputs).
 */
int  MPR121_RegisterWrite(uint8 regNum, uint8 regData)
{
    int   retVal = ERROR;

    if (I2C1MasterStart(0xB4))  // Control byte Ack'd
    {
        if (I2C1MasterSend(regNum))  // Register address Ack'd
            if (I2C1MasterSendAckStop(regData))  retVal = 0;
    }

    return retVal;
}


/*^
 * Function:     Read data from a specified register of the MPR121.
 *               I2C slave address is 0x5A; Read control byte is 0xB5.
 *
 * Entry args:   regNum  = register number (address)
 *
 * Return val:   Register data (0..0xFF) read from the MPR121 normally, or
 *               ERROR (-1) if an I2C comm fault was detected.
 */
int  MPR121_RegisterRead(uint8 regNum)
{
    int    retVal = ERROR;
    uint8  bDat = 0xFF;

    if (I2C1MasterStart(0xB4))  // Control byte Ack'd
    {
        if (I2C1MasterSend(regNum))  // Register address Ack'd
            retVal = I2C1MasterReadNackStop(0xB5, &bDat, 1);
    }

    if (retVal != ERROR) retVal = bDat;

    return retVal;
}


/*^
 * Function:     Put MPR121 in STOP condition and perform software reset.
 *
 * Return val:   OK (0) normally;  ERROR (-1) if an error is detected.
 */
int  MPR121_SoftReset()
{
    int  result = 0;

    if (MPR121_RegisterWrite(ELE_CONFIG_REG, 0) == ERROR) result = ERROR;

    if (MPR121_RegisterWrite(MPR121_RESET_REG, 0x63) == ERROR) result = ERROR;

    DelayMs(1);

    if (MPR121_RegisterRead(AFE_CONFIG2_REG) != 0x24) result = ERROR;

    return result;
}


/*^
 * Function:     Configure MPR121 electrodes for touch sense operation.
 *               The device should be reset immediately prior to calling this function.
 *               On exit, the function puts the device into "RUN" mode.
 *
 * Entry args:   electrodeCount  = Number of electrodes to be activated (1..12)
 *
 */
void  MPR121_TouchSetup(uint8 electrodeCount)
{
    short  ele;
    short  regNum = TOUCH_THRESH0_REG;

    if (electrodeCount > 12) electrodeCount = 12;

    // Put device in STOP mode
    MPR121_RegisterWrite(ELE_CONFIG_REG, 0);

#ifdef USE_SPARKFUN_CONFIG

    // Set up filter param's
    MPR121_RegisterWrite(MHD_RISING_REG, 1);
    MPR121_RegisterWrite(NHD_RISING_REG, 1);
    MPR121_RegisterWrite(NCL_RISING_REG, 0);
    MPR121_RegisterWrite(FDL_RISING_REG, 0);
    
    MPR121_RegisterWrite(MHD_FALLING_REG, 1);
    MPR121_RegisterWrite(NHD_FALLING_REG, 1);
    MPR121_RegisterWrite(NCL_FALLING_REG, 255);
    MPR121_RegisterWrite(FDL_FALLING_REG, 2);

    // Set up touch thresholds for each electrode...
    for (ele = 0;  ele < electrodeCount;  ele++)
    {
        MPR121_RegisterWrite(regNum, TOUCH_THRESHOLD);
        regNum++;
        MPR121_RegisterWrite(regNum, RELEASE_THRESHOLD);
        regNum++;
    }

    MPR121_RegisterWrite(AFE_CONFIG2_REG, 0x04);    // Set ESI b2

    // Put device in RUN mode
    MPR121_RegisterWrite(ELE_CONFIG_REG, electrodeCount);
    
#else  // assume USE_ADAFRUIT_CONFIG
    
    // Set up filter param's
    MPR121_RegisterWrite(MHD_RISING_REG, 1);
    MPR121_RegisterWrite(NHD_RISING_REG, 1);
    MPR121_RegisterWrite(NCL_RISING_REG, 14);
    MPR121_RegisterWrite(FDL_RISING_REG, 0);

    MPR121_RegisterWrite(MHD_FALLING_REG, 1);
    MPR121_RegisterWrite(NHD_FALLING_REG, 5);
    MPR121_RegisterWrite(NCL_FALLING_REG, 1);
    MPR121_RegisterWrite(FDL_FALLING_REG, 0);

    // Set up touch thresholds for each electrode...
    for (ele = 0;  ele < electrodeCount;  ele++)
    {
        MPR121_RegisterWrite(regNum, TOUCH_THRESHOLD);
        regNum++;
        MPR121_RegisterWrite(regNum, RELEASE_THRESHOLD);
        regNum++;
    }

    MPR121_RegisterWrite(AFE_CONFIG1_REG, 0x10);  // 16uA charge (POR default)
    MPR121_RegisterWrite(AFE_CONFIG2_REG, 0x20);  // 0.5uS encoding, 1ms period

    // Put device in RUN mode
    MPR121_RegisterWrite(ELE_CONFIG_REG, 0x80 | electrodeCount);

#endif
}
