/***************************************************************************************\
 * File:            HardwareProfile.h 
 *
 * Purpose:         Hardware Platform definitions common to all MJB PIC32MX projects
 *
 * Target H/W:      Olimex PIC32-MX460 Development Board 
 *                  Olimex PIC32-MX340 Prototyping Board
 *                  Custom PIC32-MX340 Board (REMI Sound Synth Module)
 *
 * Processor:       PIC32MX460F512L or PIC32MX340F512H
 *
 * Compiler:        Microchip XC32 v1.33++ under MPLAB.X v2.20++
 *
\***************************************************************************************/

#ifndef __HARDWARE_PROFILE_H
#define __HARDWARE_PROFILE_H

#include "Common/Compiler.h"

#define DEVICE_ID_PIC32MX460F512L  0x00978053
#define DEVICE_ID_PIC32MX340F512H  0x00916053
#define DEVICE_ID_MASK             0x0FFFFFFF
#define FLASH_APP_ENTRY_POINT      0x9D000000
#define GET_DEVICE_ID()            (DEVID)     // Register in MCU flash PM


//========================  Determine Target Platform ===================================
//
#ifdef __32MX460F512L__    // symbol defined in XC32 compiler 
#define USE_OLIMEX_PIC32MX460_BOARD  // REMI mk1 prototype platform 
#endif
// else assume h/w platform based on PIC32MX340F512H (eg. REMI mk2)

// Check that a valid processor type is selected in MPLAB'X Project Properties
#if !defined __32MX460F512L__ && !defined __32MX340F512H__
#error "Processor type PIC32MX460F512L or PIC32MX340F512H must be selected!"
#endif

//========================  Determine RTC Requirements ==================================
//
#define CHECK_SOFTWARE_TIMER_ACCURACY  1

// Some of MJB's PIC32 projects use an RTCC device. Select from the following.
//#define RTCC_TYPE_MCP79410
//#define RTCC_TYPE_ISL1208

// Kernel RTC software support is needed if a hardware RTCC device is used
#if defined RTCC_TYPE_ISL1208 || defined RTCC_TYPE_MCP79410
#define INCLUDE_KERNEL_RTC_SUPPORT
#endif

#if defined RTCC_TYPE_ISL1208 && defined RTCC_TYPE_MCP79410
#error "System can define only one RTCC device type!"
#endif

//========================  Select LCD Controller Type ==================================
//
#ifdef USE_OLIMEX_PIC32MX460_BOARD   // MJB's 'alpha' dev platform...
#define USE_LCD_CONTROLLER_ST7920    // Using 128x64 mono graphics LCD with ST7920 chip
#else                                // All other hardware configurations...
#define USE_LCD_CONTROLLER_KS0108    // Using 128x64 mono graphics LCD with KS0108 chip
#endif

//=================  Select Mass Storage Device and MDD File System  ====================
//
// Don't confuse this with the REMI synth -- it's a different project!
#ifdef BUILD_SOUND_FX_SYNTH_APP  
#define USE_USB_MSD_HOST_INTERFACE        // using USB flash drive
#define USE_MDD_FILE_SYSTEM               // Memory Disk Drive file system (FAT32)
#define USE_SST25VF016                    // using SPI serial data flash (SDF) memory
#define USE_SDF_FILE_SYSTEM               // using SDF file system
#endif

//========================== Select other custom hardware ===============================
//
#ifdef USE_OLIMEX_PIC32MX460_BOARD   // MJB's 'alpha' dev platform...
#define USE_4X4_KEYPAD
#ifdef BUILD_REMI_CONTROLLER_APP
#define BUILD_REMI_MK1_PROTOTYPE   // 'Rev.A' using MPR121 I2C wired handset
#endif
#endif // USE_OLIMEX_PIC32MX460_BOARD

// Assign PIC32 I2C peripheral channel(s) that the project uses...
#ifdef USE_OLIMEX_PIC32MX460_BOARD
#define USE_I2C_CHANNEL_1             // I2C1 for mk1 wired handset
#define USE_I2C_CHANNEL_2             
#define EEPROM_I2C_PORT_NUMBER   2    // I2C2 for EEPROM, etc
#else  // assume PIC32MX340 board
#define USE_I2C_CHANNEL_1             // I2C1 for all I2C devices
#define EEPROM_I2C_PORT_NUMBER   1    
#endif

//================  UART DRIVER BUILD OPTIONS  =======================
//
#define UART1_RX_INTERRUPT_DRIVEN  1   // if 0, use polled RX input
#define UART2_RX_INTERRUPT_DRIVEN  1   // if 0, use polled RX input

#define UART1_TX_USING_QUEUE  1    // if 0, use direct TX functions
#define UART2_TX_USING_QUEUE  0    // if 0, use direct TX functions

#define UART1_TXBUFSIZE      128   // TX FIFO buffer size, chars
#define UART2_TXBUFSIZE      128

#define UART1_RXBUFSIZE      256   // RX FIFO buffer size, chars
#define UART2_RXBUFSIZE      256
//
//====================================================================

// DEBUG LED driver macros (optional for applications other than REMI)
#ifdef USE_OLIMEX_PIC32MX460_BOARD
#define DEBUG_LED_INIT()    TRISDbits.TRISD1 = 0
#define DEBUG_LED_ON()      LATDbits.LATD1 = 1
#define DEBUG_LED_OFF()     LATDbits.LATD1 = 0
#define DEBUG_LED_TOGGLE()  LATDbits.LATD1 ^= 1
#else  // Debug LED not available (RD1 reassigned)
#define DEBUG_LED_INIT()    {}
#define DEBUG_LED_ON()      {}
#define DEBUG_LED_OFF()     {}
#define DEBUG_LED_TOGGLE()  {}
#endif


/*********************************************************************
* SST25VF016 SPI Flash Memory configuration:
* Define here the SPI channel and slave-select (CS#) pin to use.
*
* MJB's Olimex PIC32-MX460 dev board has an SST25VF016B interfaced 
* to SPI channel 1. SPI CS# is RB2.
*
* SPI clk rate, Fclk = FPB / (2 x (SPIxBRG + 1))
* Example:  FPB = 80MHz, SPI1BRG = 1  :: Fclk = 80 / 4 = 20 MHz
*
*********************************************************************/
#if defined USE_SST25VF016  // (Not used for REMI app)

#define SST25_SPI_CHANNEL     1   // SPI channel for SST25VF016
#define SPI_CHANNEL_1_ENABLE
#define SST25_SPI_SPIBRG      1   // Value put in SPIxBRG reg.

#define SST25_CS_TRIS     TRISBbits.TRISB2  // SPI Flash CS# = RB2
#define SST25_CS_LAT      LATBbits.LATB2

#endif // USE_SST25VF016


/*********************************************************************
* GetSystemClock() returns system clock frequency.
*
* GetPeripheralClock() returns peripheral clock frequency.
*
* GetInstructionClock() returns instruction clock frequency.
*
*********************************************************************/

// Comment out the following line if the pre-fetch cache is NOT enabled
// in the MCU core initialization sequence. (It should be enabled!)
#define CP0_PREFETCH_CACHE_ENABLED

/*********************************************************************
* Macro: #define	GetSystemClock() 
*
* Overview: This macro returns the system clock frequency in Hertz.
*
********************************************************************/

#define GetSystemClock()    (80000000ul)  
 

/*********************************************************************
* Macro: #define	GetPeripheralClock() 
*
* Overview: This macro returns the peripheral clock frequency in Hertz.
*
********************************************************************/
#define GetPeripheralClock()    (GetSystemClock() / (1 << OSCCONbits.PBDIV))


/*********************************************************************
* Macro: #define	GetInstructionClock() 
*
* Overview: This macro returns the instruction clock rate in Hertz,
*           dependent on the number of flash PM wait states configured
*           and/or CP0 kseg0 pre-fetch cache configuration.
*           (PIC32MX3xx/4xx requires 2 wait states @ SYSCLK = 80MHz.)
*
********************************************************************/
#ifdef CP0_PREFETCH_CACHE_ENABLED
#define GetInstructionClock()   (80000000)  // assume 0 wait states (cache on)
#else
#define GetInstructionClock()   ((80000000 * 10) / 23)  // assume 2 wait states
#endif

//=======================================================================================

// Debug functions used for Microchip Applications Library (MAL-ware!)
#ifdef USE_DEBUG_TRACE
////
#define DEBUG_Initialize()     // Assume debug UART iniz'd by app
#define DEBUG_PutChar(c)       putch(c)
#define DEBUG_PutString(s)     putstr(s)
#define DEBUG_PutHexUINT8(b)   putHexByte(b)
#define DEBUG_PutHexUINT16(w)  putHexWord(w)
#define DEBUG_PutHexUINT32(l)  putHexLong(l)
#define UART2PutChar(c)        putch(c)
#define UART2PrintString(s)    putstr(s)
#define UART2PutHex(b)         putHexByte(b)
#define UART2PutHexDWord(l)    putHexLong(l)
#define UART2PutDec(n)         putDecimal(n,1)
////
#else  // Debug functions not needed...
////
#define DEBUG_Initialize()
#define DEBUG_PutChar(c)
#define DEBUG_PutString(s)
#define DEBUG_PutHexUINT8(b)
#define DEBUG_PutHexUINT16(w)
#define DEBUG_PutHexUINT32(l)
#define UART2PutChar(c)
#define UART2PrintString(s)
#define UART2PutHex(b)
#define UART2PutHexDWord(l)
#define UART2PutDec(n)
////
#endif

#endif // __HARDWARE_PROFILE_H
