/*________________________________________________________________________________________*\

  File:        MDD_FS_manager.c

  Originated:  June 2014, M.J.Bauer  [www.mjbauer.biz]
  
  Overview:    Module comprises utility functions and console CLI commands for
               Memory Disk Drive (MDD) file management.
  
\*________________________________________________________________________________________*/

#include  "console.h"
#include  "RTC_support.h"
#include  "data_flash_fs.h"

#include  <ctype.h>
#include  <string.h>
#include  <stdlib.h>
#include  <math.h>

#ifdef USE_MDD_FILE_SYSTEM

extern UserSettableParameter_t  UserParam[];    // ... for "param" command
extern int   RTCC_Synchronize( void );

#if defined USE_SSD1926_SD_CARD_INTERFACE
static char msgDeviceNotMounted[] = "! SD card not mounted.\n";
#elif defined USE_USB_MSD_HOST_INTERFACE
static char msgDeviceNotMounted[] = "! USB drive not mounted.\n";
#else
static char msgDeviceNotMounted[] = "! MDD media not mounted.\n";
#endif

volatile BOOL  g_MDD_MediaMounted;     // flag: True if MDD storage device inserted


/***************************************************************************************************
 * Function: MDDmediaMountStatusUpdate()
 *
 * Overview: Periodic background task which checks for insertion and removal of MDD media.
 *           If the media was previously removed and inserted again, the FS is re-initialized.
 *           Global status flag 'g_MDD_MediaMounted' is updated accordingly.
 *
 * Warning:  System hangs (game over) if FSInit() is called without MDD media inserted!
 */
void MDDmediaMountStatusUpdate(void)
{
    static BOOL media_present_before = FALSE;
    static BOOL media_present_now;

#if defined USE_SSD1926_SD_CARD_INTERFACE
    media_present_now = SDDetect();
#elif defined USE_USB_MSD_HOST_INTERFACE
    media_present_now = USBHostMSDSCSIMediaDetect();
#else
    media_present_now = FALSE;
#endif

    if (!media_present_before && media_present_now)  // media just inserted
    {
        if (FSInit()) g_MDD_MediaMounted = TRUE;
    }
    else if (!media_present_now)
    {
        g_MDD_MediaMounted = FALSE;
    }

    media_present_before = g_MDD_MediaMounted;

    if (g_MDD_MediaMounted) UpdateFSrealTimeVars();  // for file time-stamp
}


/**
*   CLI command function:  Cmnd_format
*
*   The "format" command formats MDD storage media using the FAT32 file system.
*   All pre-existing files will be deleted, but not necessarily lost.
*/
void  Cmnd_format( int argCount, char * argValue[] )
{
#ifdef ALLOW_FORMATS

    short result;
    char  vol_name[12];

    if ( argCount == 1 || *argValue[1] == '?' )   // help wanted
    {
        putstr( "Usage:  format $$ [vol_name] \n" );
        putstr( "Format the MDD mass storage device. \n" );
        putstr( "Optional <vol_name> = media name (8 chars max). \n" );
        return;
    }

    if (!g_MDD_MediaMounted)
    {
        putstr(msgDeviceNotMounted);
        return;
    }

    if (!SuperUserAccess()) return;

    if (argCount >= 3 && strmatch(argValue[1], "$$"))
    {
        strncpy(vol_name, argValue[1], 8);  // name given
        vol_name[8] = 0;
    }
    else  strncpy(vol_name, "SD-card", 8);   // use default name

    result = FSformat(0, 0x0505, vol_name);  // arg2 is MDD serial number
    
    if (result == 0) putstr("* Media formatted OK.\n");
    else putstr("! Failed to format media.\n");
    
#else
    
    putstr("! Format disabled.\n");

#endif // ALLOW_FORMATS
}


/**
*   Local function which examines the file-name extension of a given file and 
*   if the extension is ".HEX", it is assumed to be an Intel Hex (INHX32) data file;
*   in which case the file is opened and the contents are examined for presence of a
*   'Title' record, which is assumed to be the first record of the file.
*   (The 'Title' record is a non-standard record, type == 09.)
*   If a Title record is found, the title string is copied to the given destination
*   and the function returns TRUE. Otherwise, it returns FALSE.
*/
static uint8  HexFileWithTitleRecord(char *filename, char *titleString)
{
    FSFILE *pFile;
    char    inHexRecordBuf[80];
    char   *pBuf;
    char   *extn = NULL;
    uint8   result = FALSE;
    int     len = 0;

    *titleString = 0;  // Term char (in case search fails)

    if ((extn = strchr(filename, '.')) != NULL
    && (strcmp(extn, ".HEX") == 0 || strcmp(extn, ".hex") == 0))
    {
        UpdateFSrealTimeVars();

        if ((pFile = FSfopen(filename, "r")) != NULL)
        {
            ReadNextRecord(inHexRecordBuf, pFile);  // Read first record only

            if (strncmp(&inHexRecordBuf[7], "09", 2) == 0)  // found 'title' record...
            {
                pBuf = &inHexRecordBuf[10];
                while (*pBuf != ']' && len <= 30)  // ']' is end-of-string
                {
                    if (*pBuf <= 0x20)  break;  // Ctrl char
                    *titleString++ = *pBuf++;
                    *titleString = 0;           // Term char
                    len++;
                }
                result = TRUE;
            }
            FSfclose(pFile);
         }
    }

    return result;
}


/**
*   CLI command function:  Cmnd_ls
*
*   The "ls" command lists the names of files stored on the MDD device.
*   The MDD FS uses the FAT32 format.
*   Filenames are "8.3" sized, i.e. max 12 chars (incl. dot).
*   File names are tabulated five to a line.
*   Files with ".HEX" extension are examined for presence of a 'title record';
*   if found, the 'title string' is output alongside other file info.
*/
void  Cmnd_ls( int argCount, char * argValue[] )
{
    char    filespec[16];
    char    filename_newest[32];
    char    cBuf[80];
    char    titleString[40];
    short   i, nspaces;
    short   col = 0, nfiles = 0;
    int     year, month, day, hour, mins, secs;
    BYTE    attributes = ATTR_HIDDEN | ATTR_SYSTEM | ATTR_READ_ONLY | ATTR_ARCHIVE;  // all files
    BYTE    Opt_long = 0;
    SearchRec  fileinfo;
    unsigned long timestamp = 0;  

    if ( *argValue[1] == '?' )   // help wanted
    {
        putstr( "Usage:  ls [-l] [filespec]\n" );
        putstr( "List files on MDD mass storage device. \n" );
        putstr( "<filespec> may contain wildcards '*' or '?'.\n" );
        putstr( "Option -l invokes long list format.\n" );
        return;
    }

    if (!g_MDD_MediaMounted)
    {
        putstr(msgDeviceNotMounted);
        return;
    }

    strcpy(filespec, "*.*");   // default file spec. (all files)

    if (argCount == 2)   // one arg. supplied... maybe option or filespec
    {
        if (strmatch(argValue[1], "-l")) Opt_long = 1;
        if (!Opt_long) strncpy(filespec, argValue[1], 14);
    }
    else if (argCount == 3)   // two arg's
    {
        if (strmatch(argValue[1], "-l"))  // first is option -l
        {
            Opt_long = 1;
            strncpy(filespec, argValue[2], 14);  // assume 2nd arg is filespec
        }
    }

    strcpy(filename_newest, "None matching spec.");

    while (1)
    {
        if (nfiles == 0)  // find first file
        {
            if (FindFirst(filespec, attributes, &fileinfo) == ERROR) break;
            timestamp = fileinfo.timestamp;
        }
        else  // find next file
        {
            if (FindNext(&fileinfo) == ERROR) break;
        }
        nfiles++ ;
        if (fileinfo.timestamp >= timestamp)  // found a more recent file (or same)
        {
            timestamp = fileinfo.timestamp;
            strcpy(filename_newest, fileinfo.filename);
        }
        putstr(fileinfo.filename);
        nspaces = 16 - strlen(fileinfo.filename);
        while ( col < 4 && nspaces != 0 )
        {
            putch(' ');  --nspaces;  // tab to next column
        }
        if (Opt_long)
        {
            putDecimal(fileinfo.filesize, 8);   // bytes?
            // output time-stamp  YYYY-MM-DD hh:mm
            year = (int) 1980 + (fileinfo.timestamp >> 25);
            month = (fileinfo.timestamp >> 21) & 0xF;
            day = (fileinfo.timestamp >> 16) & 0x1F;
            hour = (fileinfo.timestamp >> 11) & 0x1F;
            mins = (fileinfo.timestamp >> 5) & 0x3F;
            sprintf(cBuf, "  %4d-%02d-%02d %02d:%02d  ", year, month, day, hour, mins);
            putstr(cBuf);
            if (HexFileWithTitleRecord(fileinfo.filename, titleString))
                putstr(titleString);
            putstr("\n");
        }
        else  // short list format (file names only, tabulated 5 per line)
        {
            if ( ++col >= 5 )  { col = 0;  NEW_LINE; }
        }
    }
    if (col != 0) { NEW_LINE; }
    
    if (Opt_long)
    {
        putstr("Files matching file-spec: "); putDecimal(nfiles, 1);
        NEW_LINE;
        putstr("Most recent file created: "); putstr(filename_newest);
        NEW_LINE;
    }
}

/**
*   CLI command function:  Cmnd_rm
*
*   The "rm" command (alias "del") removes a specified file from the MDD device.
*/
void  Cmnd_rm( int argCount, char * argValue[] )
{
//  char filespec[16];
    char filename[16];

    if ( argCount == 1 || *argValue[1] == '?' )   // help wanted
    {
        putstr( "Usage:  rm  <filename> \n" );
        putstr( "Alias:  del <filename> \n" );
        putstr( "Remove (delete) file from MDD storage device.\n" );
        return;
    }

    if (!g_MDD_MediaMounted)
    {
        putstr(msgDeviceNotMounted);
        return;
    }

    strncpy(filename, argValue[1], 14);

    if (FSremove(filename) == ERROR)
    {
        putstr("! Could not remove file: ");
        putstr(filename);  NEW_LINE;
    }
}

/**
*   CLI command function:  Cmnd_view
*
*   The "view" command outputs the contents of a specified file on the MDD drive
*   to the console output stream, e.g. for viewing, or for upload to the host PC
*   using a terminal emulator "file capture" utility.
*/
void  Cmnd_view( int argCount, char * argValue[] )
{
    short result;
    FSFILE *pFile;
    char filename[16];
    BYTE bDat;
    char c;

    if ( argCount < 2 || *argValue[1] == '?' )   // help wanted
    {
        putstr( "Usage:  view  <filename> \n" );
        putstr( "Output text from file on MDD storage.\n" );
        putstr( "Hit [Esc] to terminate.\n" );
        return;
    }

    if (!g_MDD_MediaMounted)
    {
        putstr(msgDeviceNotMounted);
        return;
    }
    
    strncpy(filename, argValue[1], 14);

    pFile = FSfopen(filename, "r");
    if (pFile == NULL) 
    {
        putstr("! Could not open file: ");  putstr(filename);  NEW_LINE;
    }
    else
    {
        while (FSfread(&bDat, 1, 1, pFile) != 0)  // 0 => end of file
        {
            putch(bDat);
            if (bDat == '\n') WaitMilliseconds(20);   // Terminal can't cope with 57.6 kBaud!

            if (kbhit())
            {
                c = getch();
                if (c == ASCII_ESC) break;
            }
        }
        FSfclose(pFile);
    }
}

/**
*   CLI command function:  Cmnd_fcre8
*
*   The "fcre8" command attempts to create a new text file on the MDD storage device.
*   If the file is opened successfully, some arbitrary test data is written into the file,
*   then it is closed. The operation may be checked using the "view" command.
*/
void    Cmnd_fcre8( int argCount, char * argValue[] )
{
    static const char textBuf[] =
    {
        "Line #1: How many roads must a man walk down\r\n"
        "Line #2: Before you can call him a man?\r\n"
        "Line #3: Yes, an' how many seas must a white dove sail\r\n"
        "Line #4: Before she sleeps in the sand?\r\n"
        "Line #5: Yes, an' how many times must the cannon balls fly\r\n"
        "Line #6: Before they're forever banned?\r\n"
        "Line #7: The answer, my friend, is blowin' in the wind,\r\n"
        "Line #8: The answer is blowin' in the wind.\r\n"
    };

    short result;
    FSFILE *pFile;
    char filename[16];

    if ( argCount == 1 || *argValue[1] == '?' )   // help wanted
    {
        putstr( "Usage:  fcre8 <filename> \n" );
        putstr( "Opens a new file on the MDD storage device and \n" );
        putstr( "writes 8 lines of arbitrary text into it. \n" );
        return;
    }

    if (!g_MDD_MediaMounted)
    {
        putstr(msgDeviceNotMounted);
        return;
    }

    strncpy(filename, argValue[1], 14);

    pFile = FSfopen(filename, "w");
    if (pFile == NULL)
    {
        putstr("! Failed to open file: ");
        putstr(filename);  NEW_LINE;
    }
    else
    {
        // Write text into the file, but don't write the NUL at the end of the buffer.
        if (FSfwrite(textBuf, 1, (sizeof(textBuf)-1), pFile) != (sizeof(textBuf)-1))
            putstr("! File write failed.\n");

        FSfclose(pFile);
    }
}

#endif // USE_MDD_FILE_SYSTEM  --------------------------------------------------------------
