/*****************************************************************************
 *
 * File:      keypad_4wire.h
 *
 * Define hardware dependencies for 4-wire keypad I/O configuration here.
 *
 *****************************************************************************
 */
#ifndef _KEYPAD_4WIRE_H
#define _KEYPAD_4WIRE_H

#include "system_def.h"
#include "HardwareProfile.h"

#if defined USE_4X4_KEYPAD

// PIC32MX ADC macros
//
#ifndef ADC_INPUT_SEL

#define ADC_INPUT_SEL   AD1CHS              // ADC Input Channel Selector
#define ADC_SAMPLING    AD1CON1bits.SAMP    // ADC Sample/Hold-Start
#define ADC_CONV_DONE   AD1CON1bits.DONE    // ADC Status
#define ADC_RESULT_REG  ADC1BUF0            // ADC Conversion Result Buffer Register

#define PIN_CONFIG_ANALOG    0
#define PIN_CONFIG_DIGITAL   1

#endif

// ADC channel input assignments for keypad X and Y line voltage sensing.
// NB: X voltage is read on Y- line;  Y voltage is read on X- line.
#define ADC_CHAN_XPLUS                 (10 << 16)    // X+ = AN10
#define ADC_CHAN_XMINUS                (11 << 16)    // X- = AN11
#define ADC_CHAN_YPLUS                 (12 << 16)    // Y+ = AN12
#define ADC_CHAN_YMINUS                (13 << 16)    // Y- = AN13

#define ADPCFG_PIN_XPLUS               AD1PCFGbits.PCFG10
#define ADPCFG_PIN_XMINUS              AD1PCFGbits.PCFG11
#define ADPCFG_PIN_YPLUS               AD1PCFGbits.PCFG12
#define ADPCFG_PIN_YMINUS              AD1PCFGbits.PCFG13

// Keypad X (column) pin I/O macro definitions
#define XPLUS_DRIVE_HIGH()             LATBbits.LATB10   = 1
#define XPLUS_DRIVE_LOW()              LATBbits.LATB10   = 0
#define XPLUS_CONFIG_AS_INPUT()        TRISBbits.TRISB10 = 1
#define XPLUS_CONFIG_AS_OUTPUT()       TRISBbits.TRISB10 = 0
#define XPLUS_INPUT_STATE()            (PORTBbits.RB10)

#define XMINUS_DRIVE_HIGH()            LATBbits.LATB11   = 1
#define XMINUS_DRIVE_LOW()             LATBbits.LATB11   = 0
#define XMINUS_CONFIG_AS_INPUT()       TRISBbits.TRISB11 = 1
#define XMINUS_CONFIG_AS_OUTPUT()      TRISBbits.TRISB11 = 0

// Keypad Y (row) pin I/O macro definitions
#define YPLUS_DRIVE_HIGH()             LATBbits.LATB12   = 1
#define YPLUS_DRIVE_LOW()              LATBbits.LATB12   = 0
#define YPLUS_CONFIG_AS_INPUT()        TRISBbits.TRISB12 = 1
#define YPLUS_CONFIG_AS_OUTPUT()       TRISBbits.TRISB12 = 0
#define YPLUS_INPUT_STATE()            (PORTBbits.RB12)

#define YMINUS_DRIVE_HIGH()            LATBbits.LATB13   = 1
#define YMINUS_DRIVE_LOW()             LATBbits.LATB13   = 0
#define YMINUS_CONFIG_AS_INPUT()       TRISBbits.TRISB13 = 1
#define YMINUS_CONFIG_AS_OUTPUT()      TRISBbits.TRISB13 = 0

// Assign a spare I/O pin to use as a testpoint...
// (optional, to measure ISR cycle time on a 'scope)
#define TESTPOINT_KEYISR_HIGH()        // (LATDbits.LATD15 = 1)
#define TESTPOINT_KEYISR_LOW()         // (LATDbits.LATD15 = 0)
#define TESTPOINT_KEYISR_TOGGLE()      // (LATDbits.LATD15 ^= 1)

#define KEYPAD_TASK_PERIOD_MS   1    // Time between calls to service routine (msec)

// Time required for complete keypad reading cycle (msec)
#define ADC_SERVICE_CYCLE_TIME_MS   (5 * KEYPAD_TASK_PERIOD_MS)

// Define the "key hit" detect time...
// This value is used to adjust the keypad "de-glitch filter" response.
// The higher the value, the longer the response time, but less chance of false key hit.
#define KEY_HIT_DETECT_TIME    30    // range: 10..50, default: 30 milliseconds


enum  KeypadService_SM_States
{
    KEY_SENSE_INIT = 0,
    KEY_DOWN_SENSE,
    KEY_X_HOLD,
    KEY_X_READ,
    KEY_Y_HOLD,
    KEY_Y_READ,
};


// Functions declared below implement the application program interface (API)
// for the keypad module.
void   KeypadTaskEnable(BOOL enab);
void   KeypadService(void);

void   WaitForKeyHit(void);
uint8  KeyHit(void);
uint8  GetKey(void);
short  GetKeyPressTime_ms(void);

// Diagnostic functions
short  Keypad_GetState(void);
short  GetADCreadingKeyCol(void);
short  GetADCreadingKeyRow(void);

#endif  // USE_4X4_KEYPAD

#endif  // _KEYPAD_4WIRE_H
